import os
import pandas as pd
# import pytest
try:    import nutDb as db
except:
    print('Online Test...')
    from pyNut import nutDb as db
try:    import nutFiles as fl
except: from pyNut import nutFiles as fl



#=============================================================================
# function...
#=============================================================================
def fBl_liteAccessible(str_pathDb):
    if not fl.fBl_FolderExist(str_pathDb):
        print('You do not have access to the lite DB to test it: |{}|'.format(str_pathDb))
        return False
    return True

def fDf_lite_Launch():
    str_pathDb = r'C:\Users\laurent.tupin\Documents\GitLab\sola-pcf\db_param.db'
    if not fBl_liteAccessible(str_pathDb):
        return None
    str_req =       "SELECT * FROM tbl_connectSolaDb"
    db_lite =       db.c_db_lite()
    db_lite.definePath(str_pathDb)
    db_lite.connect()
    df_UID =        db_lite.getDataframe(str_req)
    db_lite.closeConnection()
    return df_UID



#=============================================================================
# UNIT TEST
#=============================================================================
def test_c_db_lite_singleton():
    db_lite = db.c_db_lite()
    db_lit2 = db.c_db_lite()
    db_lite.definePath('pathDb1')
    db_lit2.definePath('pathDb2')
    assert (db_lite.str_pathDb == db_lit2.str_pathDb)
    assert (db_lite.str_pathDb == 'pathDb2')
    db_lite.connect()
    assert (db_lite.cnxn is not None)
    assert (db_lit2.cnxn is not None)
    db_lit2.closeConnection()
    assert (db_lite.cnxn is None)
    assert (db_lit2.cnxn is None)

def test_c_db_sqlServer():
    dbServer = db.c_db_sqlServer()
    d_param = dict(server = '101', database = 'Test', uid = 'laurent', pwd = '**abc**')
    dbServer.defineCredentials(**d_param)
    dbServer.request = "SELECT * FROM Table"
    assert (dbServer is not None)

def test_c_db_sqlServ_single_singleton():
    dbServer = db.c_db_sqlServ_single()
    dbServe2 = db.c_db_sqlServ_single()
    d_param = dict(server='103', database='Test', uid='laurent', pwd='**abc**')
    dbServer.defineCredentials(**d_param)
    assert (dbServer.server == '103')
    assert (dbServe2.server == '103')
    d_para2 = dict(server='104', database='Prod', uid='Guillaume', pwd='**xyz**', timeout = 50, bl_AlertIfEmptyReq = False)
    dbServe2.defineCredentials(**d_para2)
    assert (dbServer.uid == 'Guillaume')
    assert (dbServe2.uid == 'Guillaume')
    assert (dbServer.timeout == 50)
    assert (dbServe2.timeout == 50)

def test_c_db_dataframeCred_singleton():
    dbServer = db.c_db_dataframeCred()
    dbServe2 = db.c_db_dataframeCred()
    d_param = dict(server='201', database='Test', uid='laurent', pwd='**abc**')
    dbServer.defineCredentials(**d_param)
    assert (dbServer.server == '201')
    assert (dbServe2.server == '201')
    d_para2 = dict(server='202', database='Prod', uid='Guillaume', pwd='**xyz**', timeout = 50, bl_AlertIfEmptyReq = False)
    dbServe2.defineCredentials(**d_para2)
    assert (dbServer.uid == 'Guillaume')
    assert (dbServe2.uid == 'Guillaume')
    assert (dbServer.timeout == 50)
    assert (dbServe2.timeout == 50)

def test_c_db_withLog_singleton():
    dbServer = db.c_db_withLog()
    dbServe2 = db.c_db_withLog()
    d_param = dict(server='301', database='Test', uid='laurent', pwd='**abc**')
    dbServer.defineCredentials(**d_param)
    assert (dbServer.server == '301')
    assert (dbServe2.server == '301')
    d_para2 = dict(server='302', database='Prod', uid='Guillaume', pwd='**xyz**', timeout = 50, bl_AlertIfEmptyReq = False)
    dbServe2.defineCredentials(**d_para2)
    assert (dbServer.uid == 'Guillaume')
    assert (dbServe2.uid == 'Guillaume')
    assert (dbServer.timeout == 50)
    assert (dbServe2.timeout == 50)



#=============================================================================
# FUNCTIONAL TEST
#=============================================================================
def test_c_db_lite_functionnal():
    str_pathDb = r'C:\Users\laurent.tupin\Documents\GitLab\sola-pcf\db_param.db'
    if not fBl_liteAccessible(str_pathDb):
        return None
    str_req = "SELECT * FROM tbl_connectSolaDb"
    db_lite = db.c_db_lite()
    db_lite.definePath(str_pathDb)
    assert(db_lite.str_pathDb == str_pathDb)
    db_lite.connect()
    assert (db_lite.cnxn is not None)
    df_UID =  db_lite.getDataframe(str_req)
    assert (isinstance(df_UID, pd.DataFrame))
    assert (isinstance(db_lite.df_req, pd.DataFrame))
    # assert (db_lite.df_req == df_UID)
    db_lite.closeConnection()
    assert (db_lite.cnxn is None)

def test_c_db_dataframeCred_dataframeCredentials():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_dataframeCred()
    dbServer.dataframeCredentials(df_UID)
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcf_reporting')

def test_c_db_withLog_dataframeCredentials():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_withLog()
    dbServer.dataframeCredentials(df_UID)
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcf_reporting')

def test_c_db_dataframeCred_change_server():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_dataframeCred()
    dbServer.dataframeCredentials(df_UID)
    # Wrong Server
    dbServer.change_server('WrongServer')
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcf_reporting')
    # right Server
    dbServer.change_server('10.229.125.101')
    assert ('10.229.125.101' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcfReporting')

def test_c_db_withLog_change_server():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_withLog()
    dbServer.dataframeCredentials(df_UID)
    # Wrong Server
    dbServer.change_server('WrongServer')
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcf_reporting')
    # right Server
    dbServer.change_server('10.228.117.59')
    assert ('10.228.117.59' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcfReporting')

def test_c_db_dataframeCred_change_database():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_dataframeCred()
    dbServer.dataframeCredentials(df_UID)
    # right Server
    dbServer.change_database('database_ppp')
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'database_ppp')
    assert (dbServer.uid == 'pcf_reporting')

def test_c_db_withLog_change_database():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_withLog()
    dbServer.dataframeCredentials(df_UID)
    # right Server
    dbServer.change_database('database_ppp')
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'database_ppp')
    assert (dbServer.uid == 'pcf_reporting')

def test_c_db_dataframeCred_connect():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_dataframeCred()
    dbServer.dataframeCredentials(df_UID)
    # Connect
    int_return = dbServer.connect()
    assert (int_return == 2)
    assert (dbServer.cnxn is not None)
    int_return = dbServer.connect()
    assert (int_return == 1)
    # New connection
    dbServer.change_server('10.229.125.101')
    int_return = dbServer.connect()
    assert (int_return == 2)
    # go back to old connection
    dbServer.change_server('D1PRDSOLADB.infocloud.local')
    int_return = dbServer.connect()
    assert (int_return == 3)
    # Close connection
    dbServer.closeConnection()
    assert (dbServer.cnxn is None)

def test_c_db_withLog_connect():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_withLog()
    dbServer.dataframeCredentials(df_UID)
    # Connect
    int_return = dbServer.connect()
    assert (int_return == 2)
    assert (dbServer.cnxn is not None)
    int_return = dbServer.connect()
    assert (int_return == 1)
    # New connection
    dbServer.change_server('10.229.125.101')
    int_return = dbServer.connect()
    assert (int_return == 2)
    # go back to old connection
    dbServer.change_server('D1PRDSOLADB.infocloud.local')
    int_return = dbServer.connect()
    assert (int_return == 3)
    # Close connection
    dbServer.closeConnection()
    assert (dbServer.cnxn is None)

def test_c_db_dataframeCred_request():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_dataframeCred()
    dbServer.dataframeCredentials(df_UID)
    dbServer.change_server('10.229.125.101')
    dbServer.change_database('SolaQC')
    dbServer.connect()
    dbServer.request = 'SELECT top 1 * FROM log ORDER BY [dtm_log]'
    dbServer.getDataframe()
    dbServer.commit()
    df_data = dbServer.df_result
    assert (df_data['str_user'].values[0] == 'laurent.tupin')
    assert (df_data['str_toolName'].values[0] == 'Sirius_PCF_dev.xlsb')

def test_c_db_withLog_request():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_withLog()
    dbServer.dataframeCredentials(df_UID)
    dbServer.change_server('10.229.125.101')
    dbServer.change_database('SolaQC')
    dbServer.connect()
    dbServer.request = 'SELECT top 1 * FROM log ORDER BY [dtm_log]'
    dbServer.getDataframe()
    dbServer.commit()
    df_data = dbServer.df_result
    assert (df_data['str_user'].values[0] == 'laurent.tupin')
    assert (df_data['str_toolName'].values[0] == 'Sirius_PCF_dev.xlsb')

def test_c_db_dataframeCred_request():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_dataframeCred()
    dbServer.dataframeCredentials(df_UID)
    dbServer.change_server('10.229.125.101')
    dbServer.change_database('SolaQC')
    dbServer.connect()
    assert (dbServer.cnxn is not None)
    assert (dbServer.cursor is not None)
    dbServer.request = 'SELECT top 1 * FROM log ORDER BY [dtm_log]'
    dbServer.executeReq()
    dbServer.commit()
    assert ('10.229.125.101' in dbServer.server)
    assert (dbServer.database == 'SolaQC')

def test_c_db_withLog_request():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_withLog()
    dbServer.dataframeCredentials(df_UID)
    dbServer.change_server('10.229.125.101')
    dbServer.change_database('SolaQC')
    dbServer.connect()
    assert (dbServer.cnxn is not None)
    assert (dbServer.cursor is not None)
    dbServer.request = 'SELECT top 1 * FROM log ORDER BY [dtm_log]'
    dbServer.executeReq()
    dbServer.commit()
    assert ('10.229.125.101' in dbServer.server)
    assert (dbServer.database == 'SolaQC')

def test_c_db_withLog_executeLog():
    df_UID = fDf_lite_Launch()
    if df_UID is None:
        return None
    dbServer = db.c_db_withLog()
    dbServer.dataframeCredentials(df_UID)
    dbServer.define_Log_Cred(str_serverForLog = '10.229.125.101', str_databaseForLog = 'SolaQC')
    dbServer.executeLog(str_logExec = 'SELECT top 1 * FROM log ORDER BY [dtm_log]')
    # Check it is still Prod Server + DB after a Log
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcf_reporting')
    dbServer.change_server('10.233.6.147')
    dbServer.executeLog(str_logExec='SELECT top 1 * FROM log ORDER BY [dtm_log]')
    # Check it is still Prod Server + DB after a Log
    assert ('10.233.6.147' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcfReporting')



#=============================================================================
# FUNCTIONAL TEST on Launching function
#=============================================================================
def launch_db_credentials():
    dbServer =  db.c_db_withLog()
    df_UID =    fDf_lite_Launch()
    dbServer.dataframeCredentials(df_UID)
    dbServer.define_Log_Cred(str_serverForLog='10.229.125.101', str_databaseForLog='SolaQC')
    return dbServer

def test_db_DefineConnectCursor():
    launch_db_credentials()
    dbServer = db.c_db_withLog()
    db.db_DefineConnectCursor('')
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcf_reporting')
    assert (dbServer.cursor is not None)
    assert (dbServer.cnxn is not None)

def test_db_EXECLog():
    launch_db_credentials()
    dbServer = db.c_db_withLog()
    dbServer.executeLog(str_logExec='SELECT top 1 * FROM log ORDER BY [dtm_log]')
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcf_reporting')
    assert (dbServer.cursor is not None)
    assert (dbServer.cnxn is not None)

def test_db_EXEC():
    launch_db_credentials()
    dbServer = db.c_db_withLog()
    db.db_EXEC('SELECT top 10 * FROM tblCountry')
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcf_reporting')
    assert (dbServer.cursor is not None)
    assert (dbServer.cnxn is not None)

def test_db_SelectReq():
    launch_db_credentials()
    dbServer = db.c_db_withLog()
    db.db_SelectReq('SELECT top 10 * FROM tblCountry')
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcf_reporting')
    assert (dbServer.cursor is not None)
    assert (dbServer.cnxn is not None)
    assert (isinstance(dbServer.df_result, pd.DataFrame))

def test_db_EXEC_dbLogInRequest():
    launch_db_credentials()
    dbServer = db.c_db_withLog()
    db.db_SelectReq('SELECT top 10 * FROM tblCountry')
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcf_reporting')
    assert (dbServer.cursor is not None)
    assert (dbServer.cnxn is not None)
    assert (isinstance(dbServer.df_result, pd.DataFrame))
    db.db_EXEC('SELECT top 1 * FROM SolaQC..log ORDER BY [dtm_log]')
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')
    assert (dbServer.uid == 'pcf_reporting')
    assert (dbServer.cursor is not None)
    assert (dbServer.cnxn is not None)
    assert (isinstance(dbServer.df_result, pd.DataFrame))

def test_getDfLog():
    launch_db_credentials()
    dbServer = db.c_db_withLog()
    db.db_SelectReq('SELECT top 10 * FROM tblCountry')
    assert ('D1PRDSOLADB' in dbServer.server)
    str_req = r"SELECT str_product FROM SolaQC..tbl_inactiveProduct WHERE str_listType = 'Restricted Stocks' AND str_perimeter = 'UK_Harvest_CSI300' AND bl_inactive = 1"
    df_sql = db.db_SelectReq(str_req, bl_AlertIfEmptyReq=True)
    assert (isinstance(df_sql, pd.DataFrame))
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (dbServer.database == 'SolaDBServer')

def test_keepSingletonInMemory():
    dbServer = db.c_db_withLog()
    assert ('D1PRDSOLADB' in dbServer.server)
    assert (isinstance(dbServer.df_UID, pd.DataFrame))
    db.db_SelectReq("SELECT top 10 * FROM tblCountry")
    assert (isinstance(dbServer.df_result, pd.DataFrame))
    dbServer.executeLog(str_logExec=r"SELECT top 1 * FROM log ORDER BY [dtm_log]")
    assert ('D1PRDSOLADB' in dbServer.server)
    db.db_SelectReq("SELECT top 5 * FROM tblCountry")
    assert (isinstance(dbServer.df_result, pd.DataFrame))




try:
    from . import _lib as lib
    from . import nutOther as oth
    from . import nutFiles as fl
except:
    try:
        import _lib as lib
        import nutOther as oth
        import nutFiles as fl
    except:
        from pyNut import _lib as lib
        from pyNut import nutOther as oth
        from pyNut import nutFiles as fl
os = lib.os()
pd = lib.pandas()
pyodbc = lib.pyodbc()
lite = lib.sqlite3()


# ---------------------------------------------------------------
# ------------- CLASS DB management -----------------------------
# ---------------------------------------------------------------
@oth.dec_singletonsClass
class c_db_lite:
    """ This class allows you to manage simple lite database"""

    def __init__(self):
        self.str_pathDb = None
        self.cnxn = None

    def definePath(self, str_pathDb):
        self.str_pathDb = str_pathDb

    def connect(self):
        cnxn = lite.connect(self.str_pathDb)
        self.cnxn = cnxn

    def getDataframe(self, str_req):
        df_req = pd.read_sql_query(str_req, self.cnxn)
        self.df_req = df_req
        return df_req

    def closeConnection(self):
        try:
            self.cnxn.close()
        except:
            pass
        self.cnxn = None

    def __del__(self):
        self.closeConnection()


class c_db_sqlServer:
    def __init__(self):
        self.__dict_Connect = {}
        self.__bl_AlertIfEmptyReq = True
        self.__timeout = 100
        self.__request = 'Request not defined'
        self.cnxn = None
        self.cursor = None
        self.__server = None
        self.__database = None
        self.__uid = None
        self.__pwd = None

    def defineCredentials(self, **d_param):
        if 'server' in d_param:
            self.__server = d_param['server']
        if 'database' in d_param:
            self.__database = d_param['database']
        if 'uid' in d_param:
            self.__uid = d_param['uid']
        if 'pwd' in d_param:
            self.__pwd = d_param['pwd']
        if 'timeout' in d_param:
            self.__timeout = d_param['timeout']
        if 'bl_AlertIfEmptyReq' in d_param:
            self.__bl_AlertIfEmptyReq = d_param['bl_AlertIfEmptyReq']

    @property
    def server(self):
        return self.__server

    @property
    def database(self):
        return self.__database

    @property
    def uid(self):
        return self.__uid

    @property
    def pwd(self):
        return self.__pwd

    @property
    def timeout(self):
        return self.__timeout

    @property
    def request(self):
        return self.__request

    @request.setter
    def request(self, str_request=''):
        if str_request != '':
            self.__request = str_request

    def __str__(self, message=''):
        str_msg = '--------------------------------------\n'
        str_msg += '  - |{}|, |{}|, |{}| \n'.format(self.__server, self.__database, self.__uid)
        str_msg += '  - |{}| \n'.format(self.__request)
        str_msg += '  - {} \n'.format(message)
        str_msg += '--------------------------------------'
        print(str_msg)
        return str_msg

    def connect(self):  # db_sqlConnectCursor
        i_cnxReturn = 0
        t_keyConnect = (self.__server, self.__database, self.__uid)
        if t_keyConnect not in self.__dict_Connect:
            # ----------------------------------------------------------------------------------
            # NEW connexion
            try:
                self.cnxn = pyodbc.connect('DRIVER={SQL Server};SERVER=' + self.__server +
                                           ';DATABASE=' + self.__database +
                                           ';UID=' + self.__uid +
                                           ';PWD=' + self.__pwd,
                                           timeout=self.__timeout)
                # trusted_connection=YES;  for Windows Authentification
                # Save the connexion in a dico with KEY = ID in tuple0
            except Exception as err:
                self.__str__(' ERROR: Your db connexion is not working || {} || {}'.format(err, str(i_cnxReturn)))
                raise
            # ----------------------------------------------------------------------------------
            self.__dict_Connect[t_keyConnect] = self.cnxn
            # Keep in mind for the next Connexion
            self.t_keyConnect_Current = t_keyConnect
            i_cnxReturn = 2
        elif self.t_keyConnect_Current != t_keyConnect:
            self.cnxn = self.__dict_Connect[t_keyConnect]
            # Keep in mind for the next Connexion
            self.t_keyConnect_Current = t_keyConnect
            i_cnxReturn = 3
        # elif self.__dict_Connect[t_keyConnect] == -1:
        #     # raise
        #     i_cnxReturn = -1
        #     return i_cnxReturn
        else:
            i_cnxReturn = 1
            return i_cnxReturn
        # CURSOR
        try:
            self.cursor = self.cnxn.cursor()
        except Exception as err:
            self.__str__(' ERROR: Your db CURSOR is not working || {} || {}'.format(err, str(i_cnxReturn)))
            raise
        # END
        return i_cnxReturn

    def availablePyodbcDrivers(self):  # db_seeDriversAvailable
        self.__drivers = pyodbc.drivers()
        return self.__drivers

    def executeReq(self):  # db_Execute
        try:
            self.cursor.execute(self.__request)
        except Exception as err:
            self.__str__(' ERROR: executeReq is not working || {}'.format(err))
            raise

    def commit(self):  # db_Commit
        try:
            self.cnxn.commit()
        except Exception as err:
            self.__str__(' ERROR: commit is not working || {}'.format(err))
            raise

    def getDataframe(self):  # getDataFrame_fReq
        self.df_result = False
        try:
            self.df_result = pd.read_sql(self.__request, self.cnxn)
            # Message if request is empty
            if self.__bl_AlertIfEmptyReq is True:
                if self.df_result.empty or self.df_result.dropna(how='all').empty:
                    self.__str__(' EMPTY: getDataframe is empty')
        except Exception as err:
            self.__str__(' ERROR: getDataframe is not working || {}'.format(err))
            raise
        return self.df_result

    def getDataframe_multipleReq(self):
        self.df_result = False
        try:
            l_resultSet = self.cursor.fetchall()
            df_resultSet = pd.DataFrame.from_records(l_resultSet)
            while (self.cursor.nextset()):
                l_resultSet = self.cursor.fetchall()
                df_resultSet_Suiv = pd.DataFrame.from_records(l_resultSet)
                df_resultSet = self.fDf_Concat(df_resultSet, df_resultSet_Suiv)
            # Final Result
            self.df_result = df_resultSet
            # Message if empty
            if self.__bl_AlertIfEmptyReq is True:
                if self.df_result.empty or self.df_result.dropna(how='all').empty:
                    self.__str__(' EMPTY: getDataframe_multipleReq is empty')
        except Exception as err:
            self.__str__(' ERROR: getDataframe_multipleReq is not working || {}'.format(err))
            raise
        return self.df_result

    @classmethod
    def fDf_Concat(cls, df1, df2):
        try:
            if len(df1.columns) >= len(df2.columns):
                df2.columns = df1.columns[:len(df2.columns)]
                df_return = pd.concat([df1, df2], ignore_index=True)
                df_return = df_return[df1.columns]
            else:
                df2.columns = list(df1.columns) + list(df2.columns[len(df1.columns):])
                df_return = pd.concat([df1, df2], ignore_index=True)
                df_return = df_return[df2.columns]
        except:
            df_return = False
        return df_return

    def closeConnection(self):
        try:
            self.__dict_Connect = {}
            self.__str__(' INFO: Your db connexion is closing !')
        except:
            pass
        try:
            self.cursor.close()
            del self.cursor
        except:
            pass
        try:
            self.cnxn.close()
        except:
            pass
        try:
            self.cnxn = None
        except:
            pass

    def __del__(self):
        self.closeConnection()


@oth.dec_singletonsClass
class c_db_sqlServ_single(c_db_sqlServer):
    """ This inehrit from c_db_sqlServer and is a singleton"""

    def __init__(self):
        super().__init__()


class c_db_dfCredentials(c_db_sqlServer):
    """ This inehrit from c_db_sqlServer and is a singleton
    Manage the Credentials in a dataframe form"""

    def __init__(self):
        super().__init__()
        self.df_UID = None

    def dataframeCredentials(self, df_UID):
        self.df_UID = df_UID
        # Get the columns of the UID given the order of columns are: Server, database, UID, Password
        l_columns = df_UID.columns
        self.serverColName = l_columns[0]
        self.databaseColName = l_columns[1]
        self.uidColName = l_columns[2]
        self.pwdColName = l_columns[3]
        try:
            str_server = df_UID[self.serverColName].values[0]
            str_database = df_UID[self.databaseColName].values[0]
            str_uid = df_UID[self.uidColName].values[0]
            str_pwd = df_UID[self.pwdColName].values[0]
        except:
            self.serverColName = 'Server'
            self.databaseColName = 'Database'
            self.uidColName = 'Uid'
            self.pwdColName = 'Password'
            str_server = df_UID[self.serverColName].values[0]
            str_database = df_UID[self.databaseColName].values[0]
            str_uid = df_UID[self.uidColName].values[0]
            str_pwd = df_UID[self.pwdColName].values[0]
        # Keep in memory the last working server
        self.__last_server = str_server
        # Set the value by the original Class
        d_param = dict(server=str_server, database=str_database, uid=str_uid, pwd=str_pwd)
        self.defineCredentials(**d_param)

    def change_server(self, str_server=''):
        if not str_server == '':
            df_UID = self.df_UID
            try:
                str_database = df_UID.loc[df_UID[self.serverColName] == str_server, self.databaseColName].values[0]
                str_uid = df_UID.loc[df_UID[self.serverColName] == str_server, self.uidColName].values[0]
                str_pwd = df_UID.loc[df_UID[self.serverColName] == str_server, self.pwdColName].values[0]
            except Exception as err:
                print('ERROR in change_server: we could not find the server: |{}| in the Dataframe provided \n'.format(
                    str_server))
                print(err)
                print(df_UID)
                return False
            # Keep in memory the last working server
            self.__last_server = str_server
            # Set the value by the original Class
            d_param = dict(server=str_server, database=str_database, uid=str_uid, pwd=str_pwd)
            self.defineCredentials(**d_param)
        return True

    def change_database(self, str_database=''):
        if not str_database == '':
            df_UID = self.df_UID
            str_server = self.__last_server
            str_uid = df_UID.loc[df_UID[self.serverColName] == str_server, self.uidColName].values[0]
            str_pwd = df_UID.loc[df_UID[self.serverColName] == str_server, self.pwdColName].values[0]
            # Set the value by the original Class
            d_param = dict(server=str_server, database=str_database, uid=str_uid, pwd=str_pwd)
            self.defineCredentials(**d_param)
        return True


@oth.dec_singletonsClass
class c_db_dataframeCred(c_db_dfCredentials):
    """ This inehrit from c_db_dfCredentials + is a singleton"""

    def __init__(self):
        super().__init__()


@oth.dec_singletonsClass
class c_db_withLog(c_db_dfCredentials):
    """ This inehrit from c_db_dfCredentials + is a singleton
    Allow you to manage some Log request inside a particular server and database
    without changing the connection of the other requests"""

    def __init__(self):
        super().__init__()
        self.__server_default = None
        self.bl_useLog = False

    def dataframeCredentials(self, df_UID):
        super().dataframeCredentials(df_UID)
        self.__server_default = self.server
        self.__database_default = self.database
        self.__uid_default = self.uid
        self.__pwd_default = self.pwd

    def change_server(self, str_server=''):
        super().change_server(str_server)
        if not str_server == '':
            self.__server_default = self.server
            self.__database_default = self.database
            self.__uid_default = self.uid
            self.__pwd_default = self.pwd

    def define_Log_Cred(self, str_serverForLog=None, str_databaseForLog=None):
        df_UID = self.df_UID
        self.server_Log = str_serverForLog
        self.db_Log = str_databaseForLog
        self.__uid_Log = df_UID.loc[df_UID[self.serverColName] == self.server_Log, self.uidColName].values[0]
        self.__pwd_Log = df_UID.loc[df_UID[self.serverColName] == self.server_Log, self.pwdColName].values[0]

    def executeLog(self, str_logExec=''):
        d_pLog = dict(server=self.server_Log, database=self.db_Log, uid=self.__uid_Log, pwd=self.__pwd_Log)
        self.defineCredentials(**d_pLog)
        self.request = str_logExec
        self.connect()
        self.executeReq()
        self.commit()
        # Go back to the original Credentials
        d_pDefault = dict(server=self.__server_default, database=self.__database_default, uid=self.__uid_default,
                          pwd=self.__pwd_default)
        self.defineCredentials(**d_pDefault)

    def getDfLog(self, str_logExec=''):
        d_pLog = dict(server=self.server_Log, database=self.db_Log, uid=self.__uid_Log, pwd=self.__pwd_Log)
        self.defineCredentials(**d_pLog)
        self.request = str_logExec
        self.connect()
        self.getDataframe()
        self.commit()
        # Go back to the original Credentials
        d_pDefault = dict(server=self.__server_default, database=self.__database_default, uid=self.__uid_default,
                          pwd=self.__pwd_default)
        self.defineCredentials(**d_pDefault)


# ---------------------------------------------------------------
# ----- Function to launch the Class ----------------------------
# ---------------------------------------------------------------
def db_defineInstance():
    dbServer = c_db_withLog()
    return dbServer


def db_DefineConnectCursor(str_req, str_server=None, str_database=None, df_UID=None, t_logId=None):
    try:
        dbServer = c_db_withLog()
        # --------------------------------
        # Should not be used - SHould be initiliazed by the App at the launch
        if not df_UID is None:
            dbServer.dataframeCredentials(df_UID)
        if not t_logId is None:
            dbServer.define_Log_Cred(str_serverForLog=t_logId(0), str_databaseForLog=t_logId(1))
        # --------------------------------

        dbServer.bl_useLog = False
        # Change the Server / database
        if not str_server is None:
            dbServer.change_server(str_server)

        # --------------------------------
        # Exception: if Database for Log is called, we need to use the server for log (LEGACY)
        #   And not change the default Database - because it cannot connect with Server
        if str_database is None:
            if '{}.'.format(dbServer.db_Log).lower() in str_req.lower():
                print('  (**) Database will be |{}| for this request: \n   **{}**'.format(dbServer.db_Log, str_req))
                dbServer.bl_useLog = True
        else:
            if dbServer.db_Log.lower() == str_database.lower():
                print('  (**) Database will be |{}| only for this reqq  \n'.format(dbServer.db_Log))
                dbServer.bl_useLog = True
            else:
                dbServer.change_database(str_database)
        # --------------------------------

        dbServer.request = str_req
        dbServer.connect()
    except Exception as err:
        print('  ERROR in db_DefineConnectCursor: {}'.format(err))
        raise
    return True


def db_SelectReq(str_req, str_server=None, str_database=None, df_UID=None, t_logId=None, bl_AlertIfEmptyReq=True):
    try:
        dbServer = c_db_withLog()
        db_DefineConnectCursor(str_req, str_server, str_database, df_UID, t_logId)
        dbServer.defineCredentials(bl_AlertIfEmptyReq=bl_AlertIfEmptyReq)
        if dbServer.bl_useLog is True:
            dbServer.getDfLog(str_req)
        else:
            dbServer.getDataframe()
            dbServer.commit()
    except Exception as err:
        print('  ERROR in db_SelectReq: {}'.format(err))
        raise
    return dbServer.df_result


def db_EXEC(str_req, str_server=None, str_database=None, df_UID=None, t_logId=None):
    try:
        dbServer = c_db_withLog()
        db_DefineConnectCursor(str_req, str_server, str_database, df_UID, t_logId)
        if dbServer.bl_useLog is True:
            dbServer.executeLog(str_req)
        else:
            dbServer.executeReq()
            dbServer.commit()
    except Exception as err:
        print('  ERROR in db_EXEC: {}'.format(err))
        raise
    return True


def db_MultipleReq(str_req, str_server=None, str_database=None, df_UID=None, t_logId=None):
    try:
        dbServer = c_db_withLog()
        db_DefineConnectCursor(str_req, str_server, str_database, df_UID, t_logId)
        dbServer.executeReq()
        dbServer.getDataframe_multipleReq()
        dbServer.commit()
    except Exception as err:
        print('  ERROR in db_MultipleReq: {}'.format(err))
        raise
    return dbServer.df_result


# -------------------------------------------------------------------------------------------------------------
# Read DB, Save into CSV and read CSV after that (avoid connexion to the same DB several times)
# -------------------------------------------------------------------------------------------------------------
@oth.dec_singletonsClass
class c_FolderCsv:
    def __init__(self):
        self.__folder = None

    @property
    def folder(self):
        return self.__folder

    @folder.setter
    def folder(self, str_folder=None):
        if not str_folder is None:
            self.__folder = str_folder


def fDf_readDB_orReadCsv(str_req, str_csvName, int_dayToKeep, str_folderCsv=None, bl_AlertIfEmptyReq=True):
    # Manage so you dont need to tell everytime the folder we are in
    if (str_folderCsv is None) or (str_folderCsv == ''):
        FolderCsv = c_FolderCsv()
        str_folderCsv = FolderCsv.folder
    else:
        FolderCsv = c_FolderCsv()
        FolderCsv.folder = str_folderCsv

    # Boolean
    bl_fileExist = False
    bl_fileTooOld = False

    # ----- Get the Path -----
    str_Path = os.path.join(str_folderCsv, str_csvName)

    # ----- Get the information : is the file exist / too old -----
    if fl.fBl_FileExist(str_Path):
        bl_fileExist = True
        if fl.fBl_fileTooOld(str_Path):
            bl_fileTooOld = True

    # ----- READ the FILE -----
    if bl_fileExist is True and bl_fileTooOld is False:
        try:
            df_return = pd.read_csv(str_Path, header=0)
            return df_return
        except Exception as err:
            print(' Warning in fDf_readDB_orReadCsv: (pd.read_csv) \n |{}| \n |{}| \n'.format(str_Path, err))
            bl_fileTooOld = True

    # ----- READ SQL information -----
    try:
        df_return = db_SelectReq(str_req, bl_AlertIfEmptyReq=bl_AlertIfEmptyReq)
    except Exception as err:
        # ----- CANT connect to SQL: Read the CSV even it is too old (AND DONT DELETE IT !!!) -----
        print(' Warning in fDf_readDB_orReadCsv (db_SelectReq): ')
        print('  ** CANT connect to SQL: Read the CSV even it is too old (AND DONT DELETE IT !!!)')
        print('  ** {}'.format(err))
        df_return = pd.read_csv(str_Path, header=0)
        return df_return

    # ----- Delete the file : too old -----
    if bl_fileExist is True and bl_fileTooOld is True:
        try:
            fl.del_fichier_ifOldEnought(str_Path, '', int_dayToKeep)
        except:
            print(' Warning in db.fDf_readDB_orReadCsv: (os.delete) \n |{}| \n'.format(str_Path))

    # ----- Save the request on CSV -----
    try:
        df_return.to_csv(str_Path, index=False, header=True)
    except:
        print(' Warning in db.fDf_readDB_orReadCsv: (df_return.to_csv) \n |{}| \n'.format(str_Path))

    return df_return



import pytest
try:
    import nutEmail as Email
except:
    print('Online Test...')
    from pyNut import nutEmail as Email


#=============================================================================
# UNIT TEST
#=============================================================================

# OUTLOOK
def test_c_email_send():
    str_from = 'sgergh'
    str_to = 'aaa@yyy'
    str_cc = 'eee@rty'
    str_bcc = 'eurv@rtyrgheryh.com'
    str_subject = 'EMAIL subghrh'
    l_pathAttach = ['re',2, None]
    bl_draft = True
    dic_param = dict(str_to=str_to, str_cc=str_cc, str_bcc=str_bcc, str_subject=str_subject, l_pathAttach=l_pathAttach,
                     str_message='', str_from=str_from, bl_draft=bl_draft)
    o_mailParam = Email.c_email(**dic_param)
    o_mailParam.paramMail_Send()
    assert o_mailParam.SentOnBehalfOfName == str_from
    assert o_mailParam.To == str_to
    assert o_mailParam.Cc == str_cc
    assert o_mailParam.Bcc == str_bcc
    assert o_mailParam.Subject == str_subject
    assert o_mailParam.l_pathAttach == l_pathAttach
    assert o_mailParam.bl_draft == bl_draft

def test_c_otlk_builder_send_param():
    str_from = 'sgeryjkgh'
    str_to = 'aaa@yyy'
    str_cc = 'eee@rty'
    str_bcc = 'eurv@rtyrgheryh.com'
    str_subject = 'EMAIL subghrh'
    l_pathAttach = ['re', 2, None]
    bl_draft = True
    dic_param = dict(str_from=str_from, str_to=str_to, str_cc=str_cc, str_bcc=str_bcc, str_subject=str_subject, l_pathAttach=l_pathAttach,
                     str_message='', bl_draft=bl_draft)
    o_buil_emailSend = Email.c_Outlook_send(**dic_param)
    o_buil_emailSend.o_emailParam.paramMail_Send()
    assert o_buil_emailSend.o_emailParam.SentOnBehalfOfName == str_from
    assert o_buil_emailSend.o_emailParam.To == str_to
    assert o_buil_emailSend.o_emailParam.Cc == str_cc
    assert o_buil_emailSend.o_emailParam.Bcc == str_bcc
    assert o_buil_emailSend.o_emailParam.Subject == str_subject
    assert o_buil_emailSend.o_emailParam.l_pathAttach == l_pathAttach
    assert o_buil_emailSend.o_emailParam.bl_draft == bl_draft

@pytest.mark.parametrize("l_folders", [([]), (['']), (['Proj - Seita']), (['Side Activities', 'Foot'])])
def test_c_email_dwld(l_folders):
    str_outAcctName = 'sola@google.com'
    str_inbox = 'Inbox'
    dic_param = dict(str_outAcctName = str_outAcctName, str_inbox = str_inbox, l_folders = l_folders)
    o_mailParam = Email.c_email(**dic_param)
    o_mailParam.paramMail_Dwld()
    assert o_mailParam.str_outAcctName == str_outAcctName
    assert o_mailParam.str_inbox == str_inbox
    assert o_mailParam.l_folders == l_folders

@pytest.mark.parametrize("l_folders", [([]), (['']), (['Proj - Seita']), (['Side Activities', 'Foot'])])
def test_c_otlk_builder_dwld_param(l_folders):
    str_outAcctName = 'sola@google.com'
    str_inbox = 'Inbox'
    dic_param = dict(bl_test = True, str_outAcctName=str_outAcctName, str_inbox = str_inbox, l_folders = l_folders)
    o_buil_emailDwld = Email.c_Outlook_dwld(**dic_param)
    o_buil_emailDwld.o_emailParam.paramMail_Dwld()
    assert o_buil_emailDwld.o_emailParam.str_outAcctName == str_outAcctName
    assert o_buil_emailDwld.o_emailParam.str_inbox == str_inbox
    assert o_buil_emailDwld.o_emailParam.l_folders == l_folders

@pytest.mark.parametrize("l_folders", [([]), (['']), (['Proj - Seita']), (['Side Activities', 'Foot'])])
def test_c_otlk_builder_dwld(l_folders):
    str_outAcctName = ''
    str_inbox = 'Inbox'
    dic_param = dict(bl_test = True, str_outAcctName=str_outAcctName, str_inbox=str_inbox, l_folders=l_folders)
    o_buil_emailDwld =  Email.c_Outlook_dwld(**dic_param)
    o_otlk_Director =   Email.c_otlk_Director(o_buil_emailDwld)
    o_otlk_Director.Download_fMail()
    assert o_otlk_Director._builder.o_emailParam.str_outAcctName == str_outAcctName
    assert o_otlk_Director._builder.o_emailParam.str_inbox == str_inbox
    assert o_otlk_Director._builder.o_emailParam.l_folders == l_folders
    o_otlk_Director._builder.getEmails()
    int_nbEmails = len(o_otlk_Director._builder.o_emails)
    assert isinstance(int_nbEmails, int)

@pytest.mark.parametrize("str_subject, str_to, str_cc, str_File_startW, str_File_endW",
                         [('Subject that should fail', '', '', 'Piece jointe',''),
                          ('Subject that should fail:', 'laurent@google.com', '', '', ''),
                          ('Subject that should fail',  'laurent@google.com', '', '', ''),
                          ('Subject that should fail',  'lthjtj@rhgtr.com', '', '', ''),
                          ('Subject that should fail', '', 'lthjtj@rhgtr.com', '', ''),
                          ('', 'Dl-MK-StAff', 'FAKE CONTACT', '', ''),
                          ('', 'FAKE CONTACT', '', '', ''),
                          ('', '', '', 'PJ that does not exist', 'Not existiting.pdf'),
                          ('Shitty Subject that should fail',  '', 'anand@googl.com', '','')])
def test_FAIL_filterRestrict(str_subject, str_to, str_cc, str_File_startW, str_File_endW):
    str_outAcctName = ''
    str_inbox = 'Inbox'
    dic_param = dict(bl_test=True, str_outAcctName=str_outAcctName, str_inbox=str_inbox, str_subject=str_subject,
                     str_to=str_to, str_cc=str_cc, str_File_startW=str_File_startW, str_File_endW=str_File_endW)
    o_buil_emailDwld =  Email.c_Outlook_dwld(**dic_param)
    o_otlk_Director =   Email.c_otlk_Director(o_buil_emailDwld)
    with pytest.raises(Exception):
        o_otlk_Director.Download_fMail()


#------------------------------------------------------------------------------------------------------
# WEBMAIL
#------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize("l_folders", [([]), (['']), (['Proj - Seita']), (['Side Activities', 'Foot'])])
def test_c_webmail_builder_dwld_param(l_folders):
    str_outAcctName = 'sola@google.com'
    str_inbox = 'Inbox'
    str_pwd = '***'
    dic_param = dict(bl_test = True, str_outAcctName=str_outAcctName, str_pwd = str_pwd, str_inbox = str_inbox, l_folders = l_folders)
    o_buil_emailDwld =  Email.c_Webmail_dwld(**dic_param)
    o_buil_emailDwld.o_emailParam.paramMail_Dwld()
    assert o_buil_emailDwld.o_emailParam.str_outAcctName == str_outAcctName
    assert o_buil_emailDwld.o_emailParam.str_pwd == str_pwd
    assert o_buil_emailDwld.o_emailParam.str_inbox == str_inbox
    assert o_buil_emailDwld.o_emailParam.l_folders == l_folders

@pytest.mark.parametrize("str_subject, str_to, str_cc, str_File_startW, str_File_endW",
                         [('Subject that should fail', '', '', 'Piece jointe',''),
                          ('Subject that should fail:',  'laurent@google.com', '', '', ''),
                          ('Subject that should fail',  'lthjtj@rhgtr.com', '', '', ''),
                          ('Subject that should fail', '', 'lthjtj@rhgtr.com', '', ''),
                          ('Subject that should fail',  '', 'anand@goo.com', '','')])
def test_FAIL_webmail_filterRestrict(str_subject, str_to, str_cc, str_File_startW, str_File_endW):
    str_outAcctName = 'sola@ihsmarkit.com'
    str_inbox = 'Inbox'
    str_pwd = 'D3lt@0n3'
    l_folders = ['PCF-Received']
    dic_param = dict(bl_test=True, str_outAcctName=str_outAcctName, str_pwd = str_pwd, str_inbox=str_inbox, str_subject=str_subject,
                     str_to=str_to, str_cc=str_cc, str_File_startW=str_File_startW, str_File_endW=str_File_endW, l_folders=l_folders)
    o_buil_emailDwld =  Email.c_Webmail_dwld(**dic_param)
    o_otlk_Director =   Email.c_otlk_Director(o_buil_emailDwld)
    with pytest.raises(Exception):
        o_otlk_Director.Download_fMail()








try:
    from . import _lib as lib
    from . import nutOther as oth
    from . import nutDate as dat
    from . import nutFiles as fl
except:
    try:
        import _lib as lib
        import nutOther as oth
        import nutDate as dat
        import nutFiles as fl
    except:
        from pyNut import _lib as lib
        from pyNut import nutOther as oth
        from pyNut import nutDate as dat
        from pyNut import nutFiles as fl
os  = lib.os()
math =lib.math()
pd  = lib.pandas()
np  = lib.numpy()
win32       = lib.win32()
pythoncom   = lib.pythoncom()
Credentials = lib.Credentials()
Account     = lib.Account()
DELEGATE    = lib.DELEGATE()
EWSTimeZone = lib.EWSTimeZone()
EWSDateTime = lib.EWSDateTime()
Configuration = lib.Configuration()
FileAttachment = lib.FileAttachment()




#=============================================================================
# DESIGN PATTERN: BUILDER for OUTLOOK / EXCHANGELIB
#=============================================================================
class c_email:
    """Product: Define the Email Param here"""
    def __init__(self, **dic_param):
        self.dic_param = dic_param
        if 'bl_test' in dic_param:  self.bl_test = dic_param['bl_test']
        else:                       self.bl_test = False
    # GET the Param
    def __str__(self):
        # str_return = ''
        # for k, v in self.dic_param.items():
        #     str_return = '\r  -{}: {}'.format(str(k), str(v))
        return str(self.dic_param)
    def paramMail_Send(self, o_mail = None):
        dic_p = self.dic_param
        try:
            # Param to be used in the Concrete Builder (just save the param...)
            if 'bl_draft' in dic_p:         self.bl_draft = dic_p['bl_draft']
            else:                           self.bl_draft = True
            if 'l_pathAttach' in dic_p:     self.l_pathAttach = dic_p['l_pathAttach']
            else:                           self.l_pathAttach = None
            if 'str_from' in dic_p:         self.SentOnBehalfOfName =   dic_p['str_from']
            if 'str_to' in dic_p:           self.To =       dic_p['str_to']
            if 'str_cc' in dic_p:           self.Cc =       dic_p['str_cc']
            if 'str_subject' in dic_p:      self.Subject =  dic_p['str_subject']
            if 'str_bcc' in dic_p:          self.Bcc =      dic_p['str_bcc']
            if 'str_message' in dic_p:      self.Body =     dic_p['str_message']
            # Change the email object itself and return it 
            if not o_mail is None:
                if 'str_from' in dic_p:     o_mail.SentOnBehalfOfName = dic_p['str_from']
                if 'str_to' in dic_p:       o_mail.To =     dic_p['str_to']
                if 'str_cc' in dic_p:       o_mail.Cc =     dic_p['str_cc']
                if 'str_subject' in dic_p:  o_mail.Subject =dic_p['str_subject']
                if 'str_bcc' in dic_p:      o_mail.Bcc =    dic_p['str_bcc']
                if 'str_message' in dic_p:  o_mail.Body =   dic_p['str_message']
        except Exception as err:
            print(' ERROR in c_email paramMail_Send: Cannot define the Params')
            print(' - ', str(err))
            raise
        return o_mail
    def paramMail_Dwld(self):
        dic_p = self.dic_param
        try:
            if 'str_outAcctName' in dic_p:  self.str_outAcctName = dic_p['str_outAcctName']
            else:                           self.str_outAcctName = ''
            if 'str_pwd' in dic_p:          self.str_pwd = dic_p['str_pwd']
            else:                           self.str_pwd = ''
            if 'str_inbox' in dic_p:        self.str_inbox = dic_p['str_inbox']
            else:                           self.str_inbox = ''
            if 'l_folders' in dic_p:        self.l_folders = dic_p['l_folders']
            else:                           self.l_folders = []
            if 'str_subject' in dic_p:      self.str_subject = dic_p['str_subject']
            else:                           self.str_subject = ''
            if 'str_emailTime' in dic_p:    self.str_emailTime = dic_p['str_emailTime']
            else:                           self.str_emailTime = ''
            if 'dte_date' in dic_p:         self.dte_date = dic_p['dte_date']
            else:                           self.dte_date = ''
            if 'str_to' in dic_p:           self.str_to = dic_p['str_to']
            else:                           self.str_to = ''
            if 'str_cc' in dic_p:           self.str_cc = dic_p['str_cc']
            else:                           self.str_cc = ''
            if 'str_File_startW' in dic_p:  self.str_File_startW = dic_p['str_File_startW']
            else:                           self.str_File_startW = ''
            if 'str_File_endW' in dic_p:    self.str_File_endW = dic_p['str_File_endW']
            else:                           self.str_File_endW = ''
            if 'str_folder' in dic_p:       self.str_folder = dic_p['str_folder']
            else:                           self.str_folder = ''
            if 'l_folder_dest' in dic_p:    self.l_folder_dest = dic_p['l_folder_dest']
            else:                           self.l_folder_dest = ''
        except Exception as err:
            print(' ERROR in c_email paramMail_Dwld: Cannot define the Params')
            print(' - ', str(err))
            raise
        return 1
#_______________________________________            


class c_otlk_Director():
    def __init__(self, builder):
        self._builder = builder 
    def SendMail(self):
        self._builder.Define_App()
        self._builder.Define_Email()
        self._builder.Attach_files()
        self._builder.Send_eMail()
    def Download_fMail(self):
        self._builder.Define_App()
        self._builder.Define_Email()
        self._builder.Dwld_docFromEmail()
    def Move_Email(self):
        self._builder.Define_App()
        self._builder.Define_Email()
        self._builder.MoveEmail()
#_______________________________________        


class c_Outlook():
    """Abstract Builder: Build the outlook object here"""
    def __init__(self, str_nameSpace = 'MAPI'):
        self.o_outlook =        None
        self.o_emails =         None
        self.o_latestMail =     None
        self.str_nameSpace =    str_nameSpace
    def Define_App(self):
        if self.o_outlook is None:
            pythoncom.CoInitialize()
            if self.str_nameSpace == '':
                self.o_outlook = win32.Dispatch('Outlook.Application')
            else:
                self.o_outlook = win32.Dispatch('Outlook.Application').GetNamespace(self.str_nameSpace)
        return self.o_outlook
    def getAccount(self, str_outAcctName = ''):
        # Default value: User email Address
        if str_outAcctName == '':
            str_outAcctName = fl.fStr_GetUserEmail('@ihsmarkit.com')
        self.str_outAcctName = str_outAcctName
        # Define the Mailbox Sola (first Folder)
        try:    
            o_Acct = self.o_outlook.Folders(self.str_outAcctName)
        except:
            str_ErrorMsg = ' ERROR in getAccount | Cannot find the Outlook Account \n'
            str_ErrorMsg += '  - Folder wanted: {} \n'.format(str_outAcctName)
            str_ErrorMsg += '  - Folder Available on this machine : \n'
            # Loop on all account user has
            for __fol in self.o_outlook.Folders:
                __fol = str(__fol)
                str_ErrorMsg += '    - {}  \n'.format(__fol)
                l_condition = [__fol.lower() == str_outAcctName.lower()]
                if [x for x in l_condition if x == True] != []:
                    self.str_outAcctName = __fol
                    break
            # Loop on all account user has
            try:    o_Acct = self.o_outlook.Folders(self.str_outAcctName)
            except Exception as err:
                print(str_ErrorMsg, '\r - ', str(err))
                raise
        self.o_Acct = o_Acct
        return o_Acct
    def getInbox(self, str_inbox = ''):
        # Default value: User email Address
        if str_inbox == '':
            str_inbox = 'Inbox'
        self.str_inbox = str_inbox
        # Define the Inbox (2nd Folder)
        try:
            o_inbox = self.o_Acct.Folders[str_inbox]
        except Exception as err:
            print(' ERROR in getInbox: Cannot find the Box: |{}|'.format(self.str_inbox))
            print(' - ', str(err))
            raise
        self.o_inbox = o_inbox
        return o_inbox
    def getFolders(self, l_folders = []):
        self.l_folders = l_folders
        try:
            o_folder = self.o_inbox
            for _fol in l_folders:
                if not _fol == '':
                    o_folder = o_folder.Folders[_fol]
        except Exception as err:
            print(' ERROR in getFolders: Cannot find the Folder: ')
            print(' - |{}|', '|'.join(l_folders))
            print(' - ', str(err))
            raise
        self.o_folder = o_folder
        return o_folder  
    #---------------------------------
    # Archive Email: PB, depends on Define_Email in (c_Outlook_dwld)
    #---------------------------------
    def MoveEmail(self):
        # Define the folder where to move the email...
        l_folder_dest = self.o_emailParam.l_folder_dest
        if l_folder_dest == []:
            return 'No Folder to move to...'
        o_folder_dest = self.getFolders_destination(l_folder_dest)
        # Move
        try:    
            self.o_latestMail.Move(o_folder_dest)
        except Exception as err:
            print(' ERROR in MoveEmail || {}'.format(str(err)))
            raise
    def getFolders_destination(self, l_folder_dest = []):
        self.l_folder_dest = l_folder_dest
        try:
            o_folder_dest = self.o_inbox
            for _fol in l_folder_dest:
                if not _fol == '':
                    o_folder_dest = o_folder_dest.Folders[_fol]
        except Exception as err:
            print(' ERROR in getFolders_destination: Cannot find the Folder: ')
            print(' - |{}|', '|'.join(l_folder_dest))
            print(' - ', str(err))
            raise
        self.o_folder_dest = o_folder_dest
        return o_folder_dest  
    #---------------------------------
    # Just for info (method should not be used for quick operation on large Email)
    #---------------------------------
    def getEmails(self):
        self.o_emails = self.o_folder.Items
    def getLastEmails(self):
        o_emails = self.o_emails.Sort("[ReceivedTime]", True)
        self.o_emails = o_emails[:50]
#_______________________________________
    

class c_Outlook_dwld(c_Outlook):
    """Concrete Builder --> provides methods to Download Docs from Emails """
    def __init__(self, **dic_param):
        self.str_nameSpace = 'MAPI'
        super(c_Outlook_dwld, self).__init__(str_nameSpace = self.str_nameSpace)
        self.o_emailParam = c_email(**dic_param)
        
    def Define_Email(self):
        self.o_emailParam.paramMail_Dwld()
        self.getAccount(self.o_emailParam.str_outAcctName)      # GET account (name@mail.com)
        self.getInbox(self.o_emailParam.str_inbox)              # GET Inbox
        self.getFolders(self.o_emailParam.l_folders)            # GET Folders
        self.getEmails_Restict()
        self.filterEmails_ToCc()
        self.filterEmails_attachName()
        self.getTheLastEmail()
        
    def Dwld_docFromEmail(self):
        # Will Allow Unit Test to stop before downloading
        if self.o_emailParam.bl_test is True:
            return True
        self.download_PJ()
        return self.bl_success
                
    def getEmails_Restict(self):
        try:
            o_folder = self.o_folder
            # Get the Param
            str_subject = self.o_emailParam.str_subject
            str_emailTime = self.o_emailParam.str_emailTime
            # No Subject Filter
            if str_subject == '':
                o_emails = o_folder.Items
            else:
                str_sql = "@SQL=(urn:schemas:httpmail:subject LIKE '%{}%')".format(str_subject)
                o_emails = o_folder.Items.Restrict(str_sql)
                if len(o_emails) == 0:
                    print('   WARNING (out): We could not find the email with the subject |{}|'.format(str_subject))
                    raise
            #---------------------------------
            # Time of the EMAIL - str_emailTime
            if not str_emailTime == '':
                d_emailTime =   eval(str_emailTime)
                str_date =      self.o_emailParam.dte_date
                str_start_time, str_end_time  = fStr_GetDtTimeBracket(str_date, d_emailTime['start'], d_emailTime['end'])                
                str_bracket =   "[ReceivedTime] >= '{}' And [ReceivedTime] <= '{}'".format(str_start_time, str_end_time)
                o_emails_time = o_emails.Restrict(str_bracket)
                if len(o_emails_time) == 0:
                    print('   WARNING: We could not find the email with the email Time |{}|'.format(str_emailTime))
                    print('     ... Subject is : |{}|'.format(str_subject))
                    print('     ... Bracket is : |{}|'.format(str_bracket)) 
                else:
                    o_emails = o_emails_time
                    print('   *|* INFO: you have filled the Param |outlook_emailTime|, which means you will restrict the email by time')
                    print('     ... str_bracket is : |{}|'.format(str_bracket))  
            #---------------------------------
        except Exception as err:
            print(' ERROR in getEmails_Restict: |{}|'.format(err))
            print(' - Email Param: ', self.o_emailParam)
            raise
        self.o_emails = o_emails
        
    def filterEmails_ToCc(self):
        try:
            o_emails = self.o_emails
            # TO
            str_to = self.o_emailParam.str_to
            if not str_to == '':
                # TODO: Meetings can make the test fails (no To in othe object than email)                
                o_emails_sub = [o_mail for o_mail in o_emails if str_to.lower() in o_mail.To.lower()]
                if len(o_emails_sub) == 0:
                    print(' ... Empty on filterEmails_ToCc with To: {} || Subject: {}'.format(str_to, self.o_emailParam.str_subject))
                    # Does not raise error in case he dont find with the To: Tolerance to that (except if no subject)
                    if self.o_emailParam.str_subject == '':
                        raise                    
                else:   o_emails = o_emails_sub
            # Cc
            str_cc = self.o_emailParam.str_cc
            if not str_cc == '':
                o_emails_sub = [o_mail for o_mail in o_emails if str_cc.lower() in o_mail.Cc.lower()]
                if len(o_emails_sub) == 0:
                    print(' ... Empty on filterEmails_ToCc with CC: {} || Subject: {}'.format(str_cc, self.o_emailParam.str_subject))
                    # Does not raise error in case he dont find with the Cc: Tolerance to that
                    if self.o_emailParam.str_subject == '':
                        raise  
                else:   o_emails = o_emails_sub
        except Exception as err:
            print(' ERROR in filterEmails: |{}|'.format(err))
            print(' - Email Param: ', self.o_emailParam)
            raise
        self.o_emails = o_emails
        
    def filterEmails_attachName(self):
        try:
            o_emails = self.o_emails
            str_File_startW = self.o_emailParam.str_File_startW
            # PJ name start
            if not str_File_startW == '':
                o_emails_sub = [o_mail for o_mail in o_emails if str_File_startW.lower() in 
                                [str(o_attach).lower()[:len(str_File_startW)] for o_attach in o_mail.Attachments]]
                if len(o_emails_sub) == 0:
                    print(' ... Empty maill list, str_File_startW: |{}|, Subject: |{}|'.format(str_File_startW, self.o_emailParam.str_subject))
                    for o_mail in o_emails[:5]:
                        print([str(o_attach) for o_attach in o_mail.Attachments])
                    raise
                else:   o_emails = o_emails_sub
            # PJ name end       
            str_File_endW = self.o_emailParam.str_File_endW
            if not str_File_endW == '':
                o_emails_sub = [o_mail for o_mail in o_emails if str_File_endW.lower() in 
                                [str(o_attach).lower()[-len(str_File_endW):] for o_attach in o_mail.Attachments]]
                if len(o_emails_sub) == 0:
                    print(' ... Empty maill list, str_File_endW: |{}|, Subject: |{}|'.format(str_File_endW, self.o_emailParam.str_subject))
                    raise
                else:   o_emails = o_emails_sub      
        except Exception as err:
            print(' ERROR in filterEmails_attachName: |{}|'.format(err))
            print(' - Email Param: ', self.o_emailParam)
            raise
        self.o_emails = o_emails
        
    def getTheLastEmail(self):
        o_emails = self.o_emails
        try:
            i_indexMail = int(np.argmax([_mail.Senton for _mail in o_emails]))
            o_latestMail = o_emails[i_indexMail]
        except Exception as err:
            print(' ERROR in getTheLastEmail: |{}|'.format(err))
            raise
        self.o_latestMail = o_latestMail
        
    def download_PJ(self):
        o_mail = self.o_latestMail
        str_File_startW = self.o_emailParam.str_File_startW
        str_File_endW = self.o_emailParam.str_File_endW
        str_folder = self.o_emailParam.str_folder
        bl_success = False
        try:
            for o_attach in o_mail.Attachments:
                str_attachName = str(o_attach).lower()
                if not str_File_startW == '':
                    if not str_File_startW.lower() == str_attachName[:len(str_File_startW)]:
                        continue # Ignore the file
                if not str_File_endW == '':
                    if not str_File_endW.lower() == str_attachName[-len(str_File_endW):]:
                        continue # Ignore the file
                o_attach.SaveAsFile(os.path.join(str_folder, str(o_attach)))
                bl_success = True
        except Exception as err:
            print(' ERROR in download_PJ: |{}|'.format(err))
            print(' - Folder: ', str_folder)
            print(' - Mail Subject: ', self.self.o_emailParam.str_subject)
            print(' - fileName: ', str_File_startW, str_File_endW)
            print(' - Email Param: ', self.o_emailParam)
            raise
        self.bl_success = bl_success
#_____________________________________________________________________
    
class c_Outlook_send(c_Outlook):
    """Concrete Builder --> provides methods to Send the Emails """
    def __init__(self, **dic_param):
        self.str_nameSpace = ''
        super(c_Outlook_send, self).__init__(str_nameSpace = self.str_nameSpace)
        self.o_emailParam = c_email(**dic_param)
        
    def Define_Email(self):
        try:
            o_mail = self.o_outlook.CreateItem(0)
            o_mail = self.o_emailParam.paramMail_Send(o_mail)
        except Exception as err:
            print(' ERROR in Define_Email (Send): |{}|'.format(err))
            print(' - Email Param: ', self.o_emailParam, '\n')
            raise
        self.o_mail = o_mail
        
    def Attach_files(self):
        self.bl_success = True
        try:
            l_pathAttach =  self.o_emailParam.l_pathAttach
            if not l_pathAttach is None:
                for attach in l_pathAttach:
                    if fl.fBl_FileExist(attach):
                        self.o_mail.Attachments.Add(attach)
                    else:
                        self.bl_success = False
                        print('{} WARNING in Attach_files'.format('\n'))
                        print(' ==> The file that could not be JOINED to the email is: |{}| \n\n'.format(str(attach)))
                        # We do not raise error as it will be only warning !!!
        except Exception as err:
            print(' ERROR in Attach_files: |{}|'.format(err))
            print(' - Email Param: ', self.o_emailParam, '\n')
            #raise
        return self.bl_success
    
    def Send_eMail(self):
        try:
            if self.o_emailParam.bl_draft is False:
                self.o_mail.Send()
            else:
                self.o_mail.Display(True)
        except Exception as err:
            print(' ERROR in Send_eMail: |{}|'.format(err))
            print(' - Email Param: ', self.o_emailParam, '\n')
            raise
        return self.bl_success
#_____________________________________________________________________


class c_Webmail():
    """Abstract Builder: Build the Webmail object here"""
    def __init__(self):
        self.o_Acct =   None
        self.o_inbox = None
        self.o_folder = None
        self.o_emails = None
        self.o_latestMail = None
        
    def Define_App(self):
        # We need the information included in c_email so we need to wait for Concrete Class to be launch
        pass
        
    def getAccount(self, str_outAcctName = '', str_pwd = ''): 
        # Default value: User email Address
        if str_outAcctName == '':
            str_outAcctName =   fl.fStr_GetUserEmail('@ihsmarkit.com')
        self.str_outAcctName =  str_outAcctName
        # Quick variables
        self.str_emailEnd =     str_outAcctName.split('@')[-1]
        str_username =          str_outAcctName
        # Define the Credentials
        o_cred = Credentials(username = str_username, password = str_pwd)
        # Define the Account
        try:
            o_Acct = Account(credentials = o_cred, primary_smtp_address = str_outAcctName, 
                             autodiscover = True, access_type = DELEGATE)
        except Exception as err:
            # print('\n INFO, we got an error (getAccount): |{}|'.format(err))
            print('\n INFO, we got an error (getAccount)')
            o_Acct = self.getAccount_wConfig(o_cred)
        self.o_Acct = o_Acct
        return self.o_Acct
        
    def getAccount_wConfig(self, o_cred):
        str_server = 'http://eumail.{}'.format(self.str_emailEnd)
        print('  *** Account did not directly, we need to define the server: |{}| in a config obj'.format(str_server))
        try:
            o_config = Configuration(credentials = o_cred, server = str_server)
            o_Acct = Account(config = o_config, primary_smtp_address = self.str_outAcctName, access_type = DELEGATE)
        except:
            print('  **** We need auth_type = NTLM')
            o_config_NTLM = Configuration(credentials = o_cred, service_endpoint = str_server, auth_type = 'NTLM')
            o_Acct = Account(config = o_config_NTLM, primary_smtp_address = self.str_outAcctName, access_type = DELEGATE)
        self.o_Acct = o_Acct
        return self.o_Acct
    
    def getInbox(self, str_inbox = ''):
        # Default value: User email Address
        if str_inbox == '':
            str_inbox = 'Inbox'
        self.str_inbox = str_inbox
        # Define the Inbox
        try:
            if str_inbox.lower() == 'inbox':
                o_inbox = self.o_Acct.inbox
            else:
                print(' ERROR in getInbox: You are not in INBOX ??!!??')
        except Exception as err:
            print(' ERROR in getInbox: Cannot find the Box: |{}|'.format(self.str_inbox))
            print(' - ', str(err))
            raise
        self.o_inbox = o_inbox
        return o_inbox
        
    def getFolders(self, l_folders = []):
        self.l_folders = l_folders
        try:
            o_folder = self.o_inbox
            for _fol in l_folders:
                if not _fol == '':
                    o_folder = o_folder / _fol
        except Exception as err:
            print(' ERROR in getFolders: Cannot find the Folder: ')
            print(' - |{}|', '|'.join(l_folders))
            print(' - ', str(err))
            raise
        self.o_folder = o_folder
        return o_folder  
    
    #---------------------------------
    # Just for info (method should not be used for quick operation on large Email)
    #---------------------------------
    def getEmails(self):
        self.o_emails = self.o_folder.all()
    def getLastEmails(self, int_nbEmails = 50):
        o_emails = self.o_emails.order_by('-datetime_received')[:int_nbEmails]
        self.o_emails = o_emails
#_______________________________________

class c_Webmail_dwld(c_Webmail):
    """Concrete Builder --> provides methods to Download Docs from Emails """
    def __init__(self, **dic_param):
        super(c_Webmail, self).__init__()
        self.o_emailParam = c_email(**dic_param)
        
    def Define_Email(self):
        self.o_emailParam.paramMail_Dwld()
        self.getAccount(self.o_emailParam.str_outAcctName, self.o_emailParam.str_pwd)
        self.getInbox(self.o_emailParam.str_inbox)              # GET Inbox
        self.getFolders(self.o_emailParam.l_folders)            # GET Folders
        self.getEmails_Restict()
        self.filterEmails_ToCc()
        self.filterEmails_attachName()
        self.getTheLastEmail()
        
    def Dwld_docFromEmail(self):
        # Will Allow Unit Test to stop before downloading
        if self.o_emailParam.bl_test is True:
            return True
        self.download_PJ()
        return self.bl_success
                
    def getEmails_Restict(self):
        try:
            o_folder = self.o_folder
            # Get the Param
            str_subject = self.o_emailParam.str_subject
            str_emailTime = self.o_emailParam.str_emailTime
            d_paramFilter = {}
            # Build the conditions
            if not str_subject == '': 
                d_paramFilter['subject__icontains'] = str_subject
            if not str_emailTime == '':
                try:
                    d_emailTime = eval(str_emailTime)
                    str_date = self.o_emailParam.dte_date
                    dt_start_time, dt_end_time = fStr_GetDtTimeBracket(str_date, d_emailTime['start'], d_emailTime['end'], bl_dateReturn = True)
                    tz = EWSTimeZone.localzone()
                    tz_start =  EWSDateTime.from_datetime(dt_start_time).astimezone(tz)
                    tz_end =    EWSDateTime.from_datetime(dt_end_time).astimezone(tz)
                    d_paramFilter['datetime_received__range'] = (tz_start, tz_end)
                except Exception as err:
                    print('  ERROR in getEmails_Restict with Time: |{}|'.format(err))
                    print('    - str_emailTime: ', str_emailTime)
                    try:    print('    - Bracket: ', tz_start, tz_end)
                    except: pass
            # Filtering
            if any(d_paramFilter):          # IF NOT EMPTY
                o_emails = o_folder.filter(**d_paramFilter)
            else:
                self.getEmails()
                o_emails = self.o_emails
            # ERROR in case its empty
            if len(list(o_emails)) == 0:
                if str_emailTime != '':
                    print(' WARNING: We could not find the email with the EMAIL time |{}|'.format(str_emailTime))
                    print('   ... with the subject |{}|'.format(str_subject))
                    try:    print('    - Bracket: ', tz_start, tz_end)
                    except: pass
                    # TODO: Look if you launch back up again the search with just SUBJECT like with OUTLOOK
                elif str_subject != '': 
                    print('   ** WARNING (exclib): We could not find the email with the subject |{}|'.format(str_subject))
                raise
        except Exception as err:
            print(' ERROR in getEmails_Restict: |{}|'.format(err))
            print(' - str_subject: ', str_subject)
            print(' - str_emailTime: ', str_emailTime)
            try:    print(' - Bracket: ', tz_start, tz_end)
            except: pass
            # print(' - Email Param: ', self.o_emailParam)
            raise
        self.o_emails = o_emails
        
    def filterEmails_ToCc(self):
        try:
            o_emails = self.o_emails
            # TO
            str_to = self.o_emailParam.str_to
            if not str_to == '':
                o_emails_sub = [o_mail for o_mail in o_emails if str_to.lower() in o_mail.display_to.lower()]
                if len(o_emails_sub) == 0:
                    print(' ... Empty on filterEmails_ToCc with To: {} || Subject: {}'.format(str_to, self.o_emailParam.str_subject))
                    # Does not raise error in case he dont find with the To: Tolerance to that (except if no subject)
                    if self.o_emailParam.str_subject == '':
                        raise                    
                else:   o_emails = o_emails_sub
            # Cc
            str_cc = self.o_emailParam.str_cc
            if not str_cc == '':
                o_emails_sub = [o_mail for o_mail in o_emails if str_cc.lower() in o_mail.display_cc.lower()]
                if len(o_emails_sub) == 0:
                    print(' ... Empty on filterEmails_ToCc with CC: {} || Subject: {}'.format(str_cc, self.o_emailParam.str_subject))
                    # Does not raise error in case he dont find with the Cc: Tolerance to that
                    if self.o_emailParam.str_subject == '':
                        raise  
                else:   o_emails = o_emails_sub
        except Exception as err:
            print(' ERROR in filterEmails: |{}|'.format(err))
            print(' - Email Param: ', self.o_emailParam)
            raise
        self.o_emails = o_emails
        
    def filterEmails_attachName(self):
        try:
            o_emails = self.o_emails
            str_File_startW = self.o_emailParam.str_File_startW
            # PJ name start
            if not str_File_startW == '':
                o_emails_sub = [o_mail for o_mail in o_emails if str_File_startW.lower() in 
                                [str(o_attach.name).lower()[:len(str_File_startW)] for o_attach in o_mail.attachments]]
                if len(o_emails_sub) == 0:
                    print(' ... Empty mail list, str_File_startW: |{}|, Subject: |{}|'.format(str_File_startW, self.o_emailParam.str_subject))
                    self.getTheLastEmail()
                    self.print_EmailInfo(self.o_latestMail)
                    raise
                else:   o_emails = o_emails_sub
            # PJ name end       
            str_File_endW = self.o_emailParam.str_File_endW
            if not str_File_endW == '':
                o_emails_sub = [o_mail for o_mail in o_emails if str_File_endW.lower() in 
                                [str(o_attach.name).lower()[-len(str_File_endW):] for o_attach in o_mail.attachments]]
                if len(o_emails_sub) == 0:
                    print(' ... Empty mail list, str_File_endW: |{}|, Subject: |{}|'.format(str_File_endW, self.o_emailParam.str_subject))
                    self.getTheLastEmail()
                    self.print_EmailInfo(self.o_latestMail)
                    raise
                else:   o_emails = o_emails_sub      
        except Exception as err:
            print(' ERROR in filterEmails_attachName: |{}|'.format(err))
            print(' - Email Param: ', self.o_emailParam)
            raise
        self.o_emails = o_emails
        
    def getTheLastEmail(self):
        o_emails = self.o_emails
        try:
            # o_emails can be a list or an object if no other filter has been applied
            if isinstance(o_emails, list):
                i_indexMail = int(np.argmax([_mail.datetime_sent for _mail in o_emails]))
                o_latestMail = o_emails[i_indexMail]
            else:
                o_latestMail = o_emails.order_by('-datetime_received')[0]
        except Exception as err:
            print(' ERROR in getTheLastEmail: |{}|'.format(err))
            raise
        self.o_latestMail = o_latestMail
        
    def print_EmailInfo(self, o_mail):
        print('   ** Last EMail: ')
        print('     - From ', o_mail.sender)
        print('     - To', o_mail.display_to)
        print('     - Cc', o_mail.display_cc)
        print('     - Date', o_mail.datetime_received)
        print('     - List of Attachements (last EMail): ', [str(o_attach.name) for o_attach in o_mail.attachments])
        
    def download_PJ(self):
        o_mail = self.o_latestMail
        str_File_startW = self.o_emailParam.str_File_startW
        str_File_endW = self.o_emailParam.str_File_endW
        str_folder = self.o_emailParam.str_folder
        bl_success = False
        try:
            for o_attach in o_mail.attachments:
                str_attachName = str(o_attach.name)
                if isinstance(o_attach, FileAttachment):
                    if not str_File_startW == '':
                        if not str_File_startW.lower() == str_attachName[:len(str_File_startW)].lower():
                            continue # Ignore the file
                    if not str_File_endW == '':
                        if not str_File_endW.lower() == str_attachName[-len(str_File_endW):].lower():
                            continue # Ignore the file
                    with open(os.path.join(str_folder, str_attachName), 'wb') as f:
                        f.write(o_attach.content)
                    bl_success = True
                else:
                    print(' WARNING in download_PJ, file is not an attachment: |{}|'.format(str_attachName))
        except Exception as err:
            print(' ERROR in download_PJ: |{}|'.format(err))
            print(' - Folder: ', str_folder)
            print(' - Mail Subject: ', self.self.o_emailParam.str_subject)
            print(' - fileName: ', str_File_startW, str_File_endW)
            print(' - Email Param: ', self.o_emailParam)
            raise
        self.bl_success = bl_success
#_____________________________________________________________________


# END DESIGN PATTERN
#=============================================================================            





#---------------------------------------------------------------
#------ Launch an email to send ------------------
#---------------------------------------------------------------  
def fBl_SendMail_desPatt(**dic_param):
    o_builder_emailSend =   c_Outlook_send(**dic_param)
    o_otlk_Director =       c_otlk_Director(o_builder_emailSend)
    o_otlk_Director.SendMail()
    bl_success = o_otlk_Director._builder.bl_success
    return bl_success


#---------------------------------------------------------------
# ------ Additional Function ------------------
#---------------------------------------------------------------  
def fStr_GetDtTimeBracket(str_date, int_start = 0, int_end = 0, str_datFormat = '%Y-%m-%d %H:%M %p', bl_dateReturn = False):
    dt_start_time = dat.fDte_formatToDate(str_date, str_dateFormat = '%Y-%m-%d', bl_stopLoop = True)
    dt_end_time = dt_start_time
    dt_start_time = dat.fDte_AddHour(dt_start_time, int_start)
    dt_end_time = dat.fDte_AddHour(dt_end_time, int_end)
    if bl_dateReturn:
        return dt_start_time, dt_end_time
    else:
        str_start_time = dt_start_time.strftime(str_datFormat)
        str_end_time = dt_end_time.strftime(str_datFormat)
    return str_start_time, str_end_time







